local L = AceLibrary("AceLocale-2.2"):new("FuBar_PerformanceFu")
local Tablet = AceLibrary("Tablet-2.0")
local Abacus = AceLibrary("Abacus-2.0")
local Crayon = AceLibrary("Crayon-2.0")

PerformanceFu = AceLibrary("AceAddon-2.0"):new("FuBarPlugin-2.0", "AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0")
local PerformanceFu = PerformanceFu

PerformanceFu.hasIcon = true

local WoW21 = UpdateAddOnMemoryUsage and true
local scriptProfile = WoW21 and GetCVar("scriptProfile") == "1"

local string_format = string.format
local table_concat = table.concat
local table_insert = table.insert

local initialMemory, currentMemory
local mem1, mem2, mem3, mem4, mem5, mem6, mem7, mem8, mem9, mem10
local timeSinceLastUpdate
local justEntered
local gcTime

local topAddons = {}
local totalAddonMemory = 0
local totalAddonCPU = 0
local new, del
do
	local list = setmetatable({},{__mode='k'})
	function new()
		local t = next(list)
		if t then
			list[t] = nil
			return t
		else
			return {}
		end
	end
	function del(t)
		for k in pairs(t) do
			t[k] = nil
		end
		list[t] = true
		return nil
	end
end

function PerformanceFu:OnInitialize()
	self.version = "2.0" .. string.sub("$Revision: 32755 $", 12, -3)
	self.date = string.sub("$Date: 2007-04-15 21:40:16 -0400 (Sun, 15 Apr 2007) $", 8, 17)

	self:RegisterDB("PerformanceFuDB")
	self:RegisterDefaults("profile", {
		framerate = true,
		latency = false,
		memory = true,
		rate = false,
		numAddons = 3,
	})
end

function PerformanceFu:OnEnable()
	initialMemory = collectgarbage('count')
	currentMemory = initialMemory
	mem1 = currentMemory
	mem2 = currentMemory
	mem3 = currentMemory
	mem4 = currentMemory
	mem5 = currentMemory
	mem6 = currentMemory
	mem7 = currentMemory
	mem8 = currentMemory
	mem9 = currentMemory
	mem10 = currentMemory
	timeSinceLastUpdate = 0
	justEntered = true
	gcTime = time()
	self:ScheduleRepeatingEvent(self.OnUpdate, 1, self)
end

local options = {
	type = "group",
	pass = true,
	get = function(key)
		if key == "scriptProfile" then
			return GetCVar("scriptProfile") == "1"
		end
		return PerformanceFu.db.profile[key]
	end,
	set = function(key, value)
		if key == "scriptProfile" then
			SetCVar("scriptProfile", value and "1" or "0")
			ReloadUI()
		end
		PerformanceFu.db.profile[key] = value
		PerformanceFu:Update()
	end,
	func = "OnClick",
	args = {
		framerate = {
			type = 'toggle',
			name = L["Show framerate"],
			desc = L["Toggle whether to framerate"],
		},
		latency = {
			type = 'toggle',
			name = L["Show latency"],
			desc = L["Toggle whether to latency (lag)"],
		},
		memory = {
			type = 'toggle',
			name = L["Show memory usage"],
			desc = L["Toggle whether to show current memory usage"],
		},
		rate = {
			type = 'toggle',
			name = L["Show rate of increasing memory usage"],
			desc = L["Toggle whether to show increasing rate of memory"],
		},
		numAddons = {
			type = 'range',
			name = L["Number of addons"],
			desc = L["Set the number of addons to show memory usage for"],
			min = 1,
			max = 20,
			step = 1,
			hidden = function() return not WoW21 end,
		},
		scriptProfile = {
			type = 'toggle',
			name = L["Enable CPU profiling"],
			desc = L["Enable CPU profiling. Note: Reloads UI"],
			hidden = function() return not WoW21 end,
		},
		forcegc = {
			type = 'execute',
			name = L["Force garbage collection"],
			desc = L["Force a garbage collection to happen"],
			order = 101,
			func = collectgarbage
		},
	}
}
PerformanceFu.OnMenuRequest = options

local function mySort(alpha, bravo)
	if not alpha then
		return false
	elseif not bravo then
		return true
	end
	if alpha.cpu == bravo.cpu then
		return alpha.mem > bravo.mem
	else
		return alpha.cpu > bravo.cpu
	end
end

function PerformanceFu:OnUpdate()
	if not timeSinceLastUpdate then
		timeSinceLastUpdate = 0
	end
	timeSinceLastUpdate = timeSinceLastUpdate + 1

	if WoW21 and timeSinceLastUpdate >= 10 then
		for i, v in ipairs(topAddons) do
			topAddons[i] = del(v)
		end
		UpdateAddOnMemoryUsage()
		if scriptProfile then
			UpdateAddOnCPUUsage()
		end
		totalAddonMemory = 0
		totalAddonCPU = 0
		
		local num = GetNumAddOns()
		for i = 1, num do
			local v = new()
			topAddons[i] = v
			v.name = GetAddOnInfo(i)
			local mem = GetAddOnMemoryUsage(i)
			v.mem = mem
			totalAddonMemory = totalAddonMemory + mem
			if scriptProfile then
				local cpu = GetAddOnCPUUsage(i)
				v.cpu = cpu
				totalAddonCPU = totalAddonCPU + cpu
			end
		end
		table.sort(topAddons, mySort)
		
		for i = self.db.profile.numAddons+1, num do
			topAddons[i] = del(topAddons[i])
		end
	end

	if justEntered then
		if timeSinceLastUpdate >= 10 then
			initialMemory = collectgarbage('count')
			currentMemory = initialMemory
			mem1 = currentMemory
			mem2 = currentMemory
			mem3 = currentMemory
			mem4 = currentMemory
			mem5 = currentMemory
			mem6 = currentMemory
			mem7 = currentMemory
			mem8 = currentMemory
			mem9 = currentMemory
			mem10 = currentMemory
			justEntered = false
		end
	else
		mem1, mem2, mem3, mem4, mem5, mem6, mem7, mem8, mem9, mem10 =
			currentMemory, mem1, mem2, mem3, mem4, mem5, mem6, mem7, mem8, mem9
		currentMemory = collectgarbage('count')
		self:Update()
	end

	if timeSinceLastUpdate >= 10 then
		timeSinceLastUpdate = nil
	end
end

local t = {}
function PerformanceFu:OnTextUpdate()
	if not mem10 then
		mem10 = currentMemory
	end
	local currentRate = (currentMemory - mem10) / 10

	if self.db.profile.framerate then
		local framerate = floor(GetFramerate() + 0.5)
		table_insert(t, format("|cff%s%d|r fps", Crayon:GetThresholdHexColor(framerate / 60), framerate))
	end
	if self.db.profile.latency then
		local latency = select(3, GetNetStats())
		table_insert(t, format("|cff%s%d|r ms", Crayon:GetThresholdHexColor(latency, 1000, 500, 250, 100, 0), latency))
	end
	if self.db.profile.memory then
		table_insert(t, format("|cff%s%.1f|r MiB", Crayon:GetThresholdHexColor(currentMemory, 51200, 40960, 30520, 20480, 10240), currentMemory / 1024))
	end
	if self.db.profile.rate then
		table_insert(t, format("|cff%s%.1f|r KiB/s", Crayon:GetThresholdHexColor(currentRate, 30, 10, 3, 1, 0), currentRate))
	end

	self:SetText(table_concat(t, " "))

	for i = 1, #t do
		t[i] = nil
	end
end

function PerformanceFu:OnTooltipUpdate()
	local cat = Tablet:AddCategory(
		'columns', 2,
		'child_textR', 1,
		'child_textG', 1,
		'child_textB', 0
	)
	local framerate = GetFramerate()
	local r, g, b = Crayon:GetThresholdColor(framerate / 60)
	cat:AddLine(
		'text', L["Framerate:"],
		'text2', string_format("%.1f fps", framerate),
		'text2R', r,
		'text2G', g,
		'text2B', b
	)

	cat = Tablet:AddCategory(
		'text', L["Network status"],
		'columns', 2,
		'child_textR', 1,
		'child_textG', 1,
		'child_textB', 0,
		'child_text2R', 1,
		'child_text2G', 1,
		'child_text2B', 1
	)
	local bandwidthIn, bandwidthOut, latency = GetNetStats()
	bandwidthIn, bandwidthOut = bandwidthIn * 1024, bandwidthOut * 1024
	local r, g, b = Crayon:GetThresholdColor(latency, 3000, 1000, 250, 100, 0)
	cat:AddLine(
		'text', L["Latency:"],
		'text2', string_format("%d ms", latency),
		'text2R', r,
		'text2G', g,
		'text2B', b
	)

	cat:AddLine(
		'text', L["Bandwidth in:"],
		'text2', string_format("%d B/s", bandwidthIn)
	)

	cat:AddLine(
		'text', L["Bandwidth out:"],
		'text2', string_format("%d B/s", bandwidthOut)
	)

	if WoW21 then
		cat = Tablet:AddCategory(
			'text', scriptProfile and L["Addon memory/CPU usage"] or L["Addon memory usage"],
			'columns', scriptProfile and 3 or 2,
			'child_textR', 1,
			'child_textG', 1,
			'child_textB', 0,
			'child_text2R', 1,
			'child_text2G', 1,
			'child_text2B', 1,
			'child_text3R', 1,
			'child_text3G', 1,
			'child_text3B', 1
		)

		for i, v in ipairs(topAddons) do
			if v.mem == 0 then break end
			local mem = v.mem
			local text2
			if mem > 1024 then
				text2 = string_format("%.1f MiB", mem / 1024)
			else
				text2 = string_format("%.1f KiB", mem)
			end
			local cpu = v.cpu
			local text3
			if cpu then
				if cpu > 1000 then
					text3 = string_format("%.3f s", cpu/1000)
				else
					text3 = string_format("%.3f ms", cpu)
				end
			end
			cat:AddLine(
				'text', v.name,
				'text2', text2,
				'text3', text3
			)
		end
	end

	local averageRate = (currentMemory - initialMemory) / (time() - gcTime)
	cat = Tablet:AddCategory(
		'text', L["Memory usage"],
		'columns', 2,
		'child_textR', 1,
		'child_textG', 1,
		'child_textB', 0,
		'child_text2R', 1,
		'child_text2G', 1,
		'child_text2B', 1
	)
	local currentRate = (currentMemory - mem10) / 10

	local r, g, b = Crayon:GetThresholdColor(currentMemory, 51200, 40960, 30520, 20480, 10240)
	cat:AddLine(
		'text', L["Current memory:"],
		'text2', string_format("%.1f MiB", currentMemory / 1024),
		'text2R', r,
		'text2G', g,
		'text2B', b
	)

	r, g, b = Crayon:GetThresholdColor(initialMemory, 51200, 40960, 30520, 20480, 10240)
	cat:AddLine(
		'text', L["Initial memory:"],
		'text2', string_format("%.1f MiB", initialMemory / 1024),
		'text2R', r,
		'text2G', g,
		'text2B', b
	)

	if totalAddonMemory > 0 then
		r, g, b = Crayon:GetThresholdColor(totalAddonMemory, 51200, 40960, 30520, 20480, 10240)
		cat:AddLine(
			'text', L["Total addon memory:"],
			'text2', string_format("%.1f MiB", totalAddonMemory / 1024),
			'text2R', r,
			'text2G', g,
			'text2B', b
		)
	end
	if totalAddonCPU > 0 then
		local text2
		if totalAddonCPU > 1000 then
			text2 = string_format("%.3f s", totalAddonCPU/1000)
		else
			text2 = string_format("%.3f ms", totalAddonCPU)
		end
		cat:AddLine(
			'text', L["Total addon CPU usage:"],
			'text2', text2
		)
	end

	r, g, b = Crayon:GetThresholdColor(currentRate, 30, 10, 3, 1, 0)
	cat:AddLine(
		'text', L["Increasing rate:"],
		'text2', string_format("%.1f KiB/s", currentRate),
		'text2R', r,
		'text2G', g,
		'text2B', b
	)

	r, g, b = Crayon:GetThresholdColor(averageRate, 30, 10, 3, 1, 0)
	cat:AddLine(
		'text', L["Average increasing rate:"],
		'text2', format("%.1f KiB/s", averageRate),
		'text2R', r,
		'text2G', g,
		'text2B', b
	)

	Tablet:SetHint(L["Hint"])
end

function PerformanceFu:OnClick()
	collectgarbage('collect')
end

